from django.db.models import Sum
from django.http.response import Http404
from django.shortcuts import get_object_or_404
from django.core.exceptions import PermissionDenied
from django.conf import settings
from rest_framework.generics import \
    ListAPIView, ListCreateAPIView, CreateAPIView, RetrieveUpdateDestroyAPIView, RetrieveAPIView
from rest_framework.views import APIView
from rest_framework.parsers import MultiPartParser, FileUploadParser
from rest_framework.permissions import IsAuthenticated, IsAdminUser
from rest_framework.response import Response
from rest_framework import status
from rest_framework.exceptions import ValidationError
from arkindex.project.mixins import CorpusACLMixin
from arkindex.documents.models import Corpus, Right, Element, ElementType
from arkindex.dataimport.models import \
    DataImport, DataFile, DataImportState, DataImportMode, DataImportFailure, Repository
from arkindex.dataimport.serializers import \
    DataImportLightSerializer, DataImportSerializer, DataImportFailureSerializer, DataFileSerializer, \
    RepositoryLightSerializer, RepositorySerializer, ExternalRepositorySerializer
from arkindex.users.models import OAuthCredentials
import hashlib
import magic


class DataImportsList(CorpusACLMixin, ListCreateAPIView):
    """
    List data imports
    """
    permission_classes = (IsAuthenticated, )
    serializer_class = DataImportLightSerializer

    def get_queryset(self):
        if 'corpus' in self.request.query_params:
            return DataImport.objects.filter(
                corpus=self.get_corpus(self.request.query_params['corpus'])
            )
        return DataImport.objects.filter(creator=self.request.user)

    def perform_create(self, serializer):
        # Temporary limitation
        if serializer.validated_data['mode'] not in (DataImportMode.Images, ):
            raise ValidationError('Unsupported mode for now, sorry.')

        if not self.has_write_access(serializer.validated_data['corpus']):
            raise PermissionDenied

        return super().perform_create(serializer)


class DataImportDetails(CorpusACLMixin, RetrieveUpdateDestroyAPIView):
    """
    Retrieve and edit a data import
    """

    permission_classes = (IsAuthenticated, )
    serializer_class = DataImportSerializer

    def get_queryset(self):
        return DataImport.objects.filter(corpus__in=Corpus.objects.readable(self.request.user))

    def perform_update(self, serializer):
        dataimport = serializer.instance
        if not self.has_write_access(dataimport.corpus):
            raise PermissionDenied

        if dataimport.state not in (DataImportState.Created, DataImportState.Configured):
            raise ValidationError("Cannot update a workflow after it has been started")

        # Allow updating the state from Configured to Running only
        if dataimport.state != serializer.validated_data['state']:
            if dataimport.state != DataImportState.Configured:
                raise ValidationError({'state': ['Cannot change state on non-configured or started workflows']})
            if serializer.validated_data['state'] != DataImportState.Running:
                raise ValidationError(
                    {'state': ['The state of a configured workflow can only be set to configured or running']})
            dataimport.start()
            return

        super().perform_update(serializer)

        # Automatically change from Created to Configured if a payload has been set
        if dataimport.state == DataImportState.Created and dataimport.payload is not None:
            dataimport.state = DataImportState.Configured
            dataimport.save()

    def perform_destroy(self, instance):
        if not self.has_write_access(instance.corpus):
            raise PermissionDenied
        if instance.state == DataImportState.Running:
            raise ValidationError("Cannot delete a workflow while it is running")
        super().perform_destroy(instance)


class DataImportFailures(ListAPIView):
    """
    List a single import's errors
    """

    permission_classes = (IsAuthenticated, )
    serializer_class = DataImportFailureSerializer

    def get_queryset(self):
        return DataImportFailure.objects.filter(
            dataimport__corpus__in=Corpus.objects.readable(self.request.user),
            dataimport_id=self.kwargs['pk']
        ).prefetch_related('dataimport__revision__repo', 'element').order_by('path', 'line')


class DataImportDemo(CorpusACLMixin, CreateAPIView):
    """
    Create, configure and start an Images workflow from a single DataFile
    """

    permission_classes = (IsAuthenticated, )
    serializer_class = DataImportSerializer

    def create(self, request, pk=None, **kwargs):

        # Get Datafile
        corpora = Corpus.objects.writable(self.request.user)
        datafile = get_object_or_404(DataFile, corpus__in=corpora, id=pk)

        # Get volume, defaults to demo one
        volume_id = self.request.query_params.get('volume')
        filters = {
            'corpus__in': corpora,
            'type': ElementType.Volume,
        }
        if volume_id:
            filters['id'] = volume_id
        else:
            filters['name'] = 'Demo'

        volume = get_object_or_404(Element, **filters)

        assert self.has_write_access(volume.corpus), \
            'Corpus is not writable'

        # Start the import
        dataimport = DataImport.objects.create(
            creator=self.request.user,
            corpus=datafile.corpus,
            state=DataImportState.Configured,
            mode=DataImportMode.Images,
            payload={
                'volume_id': str(volume.id),
                'folder_name': 'user-{}-demo'.format(self.request.user.id),
            },
        )
        dataimport.files.add(datafile)
        dataimport.start()

        return Response(data=self.get_serializer(dataimport).data, status=status.HTTP_201_CREATED)


class DataFileList(CorpusACLMixin, ListAPIView):
    """
    List uploaded files
    """
    permission_classes = (IsAuthenticated, )
    serializer_class = DataFileSerializer

    def get_queryset(self):
        return DataFile.objects.filter(corpus=self.get_corpus(self.kwargs['pk']))


class DataFileRetrieve(CorpusACLMixin, RetrieveUpdateDestroyAPIView):
    """
    Get one file
    """
    permission_classes = (IsAuthenticated, )
    serializer_class = DataFileSerializer

    def get_queryset(self):
        return DataFile.objects.filter(corpus__in=Corpus.objects.readable(self.request.user))

    def perform_update(self, serializer):
        if not self.has_write_access(serializer.instance.corpus):
            raise PermissionDenied
        return super().perform_update(serializer)

    def perform_destroy(self, instance):
        if not self.has_write_access(instance.corpus):
            raise PermissionDenied
        return super().perform_destroy(instance)


class DataFileUpload(CorpusACLMixin, APIView):
    """
    Upload a new file to a corpus
    """
    permission_classes = (IsAuthenticated, )
    parser_classes = (MultiPartParser, FileUploadParser)

    def post(self, request, pk=None, format=None):
        if 'file' not in request.FILES:
            raise ValidationError({'file': ['No file was sent in the request']})
        try:
            corpus = self.get_corpus(pk, right=Right.Write)
        except Http404:
            raise ValidationError({'corpus': ['Corpus not found']})

        file_obj = request.FILES['file']

        if self.request.user.groups.filter(name="Demo").exists() and \
                corpus.files.exists() and \
                corpus.files.aggregate(Sum('size'))['size__sum'] + file_obj.size > settings.DEMO_FILES_LIMIT:
            raise ValidationError("Corpus total file size limit exceeded")

        md5hash = hashlib.md5()
        for chunk in file_obj.chunks():
            md5hash.update(chunk)
        file_hash = md5hash.hexdigest()

        # Reopen file to reread from beginning
        file_obj.open()
        file_type = magic.from_buffer(file_obj.read(1024), mime=True)

        if DataFile.objects.filter(hash=file_hash, corpus=corpus).exists():
            raise ValidationError('File already exists')

        df = DataFile.objects.create(
            hash=file_hash,
            corpus=corpus,
            name=file_obj.name,
            size=file_obj.size,
            content_type=file_type,
        )

        file_obj.open()
        with open(df.staging_path, 'wb') as destfile:
            for chunk in file_obj.chunks():
                destfile.write(chunk)

        return Response(
            data=DataFileSerializer(df).data,
            status=status.HTTP_201_CREATED)


class GitRepositoryImportHook(APIView):
    """
    Handle Git push events
    """

    def post(self, request, pk=None, **kwargs):
        repo = get_object_or_404(Repository, id=pk)
        repo.provider_class(credentials=repo.credentials).handle_webhook(repo, request)
        return Response(status=status.HTTP_204_NO_CONTENT)


class RepositoryList(ListAPIView):
    permission_classes = (IsAuthenticated, )
    serializer_class = RepositoryLightSerializer

    def get_queryset(self):
        return Repository.objects.filter(
            credentials__user=self.request.user,
            corpus__in=Corpus.objects.readable(self.request.user),
        )


class AvailableRepositoriesList(ListCreateAPIView):
    permission_classes = (IsAuthenticated, )
    pagination_class = None
    serializer_class = ExternalRepositorySerializer

    def get_queryset(self):
        cred = get_object_or_404(OAuthCredentials, user=self.request.user, id=self.kwargs['pk'])
        return cred.git_provider_class(credentials=cred).list_repos(
            query=self.request.query_params.get('search'),
        )

    def create(self, request, *args, **kwargs):
        serializer = self.get_serializer(data=request.data)
        serializer.is_valid(raise_exception=True)
        cred = get_object_or_404(OAuthCredentials, user=self.request.user, id=self.kwargs['pk'])

        provider = cred.git_provider_class(credentials=cred)
        repo = provider.create_repo(request=self.request, **serializer.validated_data)
        rev, _ = provider.get_or_create_latest_revision(repo)
        dataimport = rev.start_import()

        return Response(data={'import_id': str(dataimport.id)}, status=status.HTTP_201_CREATED)


class RepositoryRetrieve(CorpusACLMixin, RetrieveUpdateDestroyAPIView):
    permission_classes = (IsAuthenticated, )
    serializer_class = RepositorySerializer

    def get_queryset(self):
        return Repository.objects.filter(
            credentials__user=self.request.user,
            corpus__in=Corpus.objects.readable(self.request.user),
        )

    def perform_update(self, serializer):
        if not self.has_write_access(serializer.instance.corpus):
            raise PermissionDenied
        return super().perform_update(self, serializer)

    def perform_destroy(self, instance):
        if not self.has_write_access(instance.corpus):
            raise PermissionDenied
        return super().perform_destroy(self, instance)


class RepositoryStartImport(RetrieveAPIView):
    permission_classes = (IsAdminUser, )

    def get_queryset(self):
        return Repository.objects.filter(
            credentials__user=self.request.user,
            corpus__in=Corpus.objects.writable(self.request.user),
        )

    def get(self, request, *args, **kwargs):
        repo = self.get_object()

        rev, _ = repo.credentials.git_provider_class(
            credentials=repo.credentials,
        ).get_or_create_latest_revision(repo)

        if rev.dataimports.filter(state=DataImportState.Running).exists():
            raise ValidationError("An import is already running for the latest revision")

        return Response(data={'import_id': str(rev.start_import().id)})
