from django.db import models
from django.contrib.postgres.fields import JSONField
from django.template.defaultfilters import slugify
from django.conf import settings
from celery import states
from celery.canvas import Signature
from celery.result import AsyncResult, GroupResult
from enumfields import EnumField, Enum
from arkindex.project.celery import app as celery_app
from arkindex.dataimport.providers import git_providers, get_provider
from arkindex.project.models import IndexableModel
from arkindex.project.fields import ArrayField
import uuid
import os
import re


class DataImportState(Enum):
    Created = 'created'
    Configured = 'configured'
    Running = 'running'
    Done = 'done'
    Error = 'error'


class DataImportMode(Enum):
    Images = 'images'
    Annotations = 'annotations'
    Surfaces = 'surfaces'
    Acts = 'acts'
    Repository = 'repository'


class DataImport(IndexableModel):
    """
    A single import workflow
    """

    creator = models.ForeignKey('users.User', on_delete=models.CASCADE, related_name='imports')
    corpus = models.ForeignKey('documents.Corpus', on_delete=models.CASCADE, related_name='imports')
    state = EnumField(DataImportState, default=DataImportState.Created, max_length=30)
    mode = EnumField(DataImportMode, max_length=30)
    files = models.ManyToManyField('dataimport.DataFile', related_name='imports')
    revision = models.ForeignKey(
        'dataimport.Revision', related_name='dataimports', on_delete=models.CASCADE, blank=True, null=True)
    payload = JSONField(null=True, blank=True)
    root_id = models.UUIDField(null=True, blank=True)
    task_count = models.PositiveSmallIntegerField(null=True, blank=True)

    class Meta:
        ordering = ['corpus', '-created']

    @property
    def celery_root(self):
        return AsyncResult(str(self.root_id)) if self.root_id else None

    @property
    def tasks(self):
        """
        Get a list of all tasks associated with the import
        """
        return list(filter(
            lambda task: not isinstance(task, GroupResult),
            reversed(self.celery_root.build_graph(True).topsort())
        )) if self.celery_root else []

    @property
    def result(self):
        """
        Get the last task's result
        """
        if self.tasks and self.tasks[-1].state == states.SUCCESS:
            return self.tasks[-1].result

    def build_workflow(self):
        if self.mode == DataImportMode.Images:
            from arkindex.dataimport.tasks import check_images, import_images, save_ml_results
            analyze = celery_app.signature('arkindex_ml.tasks.analyze_pages')
            workflow = check_images.s(self) | import_images.s(self) | analyze
            workflow |= save_ml_results.s(dataimport_id=self.id)
            return workflow

        elif self.mode == DataImportMode.Repository:
            from arkindex.dataimport.tasks import download_repo, import_repo, cleanup_repo
            return download_repo.si(self) | import_repo.si(self) | cleanup_repo.si(self)

        else:
            raise NotImplementedError

    def get_task_count(self, signature):
        assert isinstance(signature, Signature)

        def _size(sig):
            # Recursive calc
            out = 0
            if hasattr(sig, 'body'):
                out += _size(sig.body)
            if hasattr(sig, 'tasks'):
                out += sum(map(_size, sig.tasks))
            else:
                out = 1
            return out

        return _size(signature)

    def start(self):
        assert self.state != DataImportState.Running
        signature = self.build_workflow()

        parent_task = signature.delay()
        # Loop to get the parent task ID (root ID)
        while parent_task.parent is not None:
            parent_task = parent_task.parent
        self.root_id = parent_task.id

        self.task_count = self.get_task_count(signature)
        self.state = DataImportState.Running
        self.save()

    @property
    def folder_name(self):
        if 'folder_name' not in self.payload:
            return
        return slugify(self.payload['folder_name'])

    @property
    def iiif_path(self):
        if not self.folder_name:
            return
        return os.path.join(settings.LOCAL_MEDIA_ROOT, self.folder_name)

    def save_failures(self, failures):
        assert self.revision is not None
        assert all(isinstance(failure, DataImportFailure) for failure in failures)
        for failure in failures:
            failure.path = re.sub(self.revision.repo.clone_dir, '', failure.path, count=1)
            failure.dataimport = self
        DataImportFailure.objects.bulk_create(failures)


class DataImportFailure(models.Model):
    id = models.UUIDField(primary_key=True, default=uuid.uuid4)
    dataimport = models.ForeignKey(
        'dataimport.DataImport', on_delete=models.CASCADE, related_name='failures')
    element = models.ForeignKey(
        'documents.Element', on_delete=models.CASCADE, blank=True, null=True, related_name='failures')
    path = models.CharField(max_length=250)
    line = models.PositiveIntegerField(blank=True, null=True)
    message = models.TextField()
    context = models.TextField(blank=True, null=True)

    @property
    def view_url(self):
        if not self.line or not self.dataimport.revision:
            return None
        return '{}/blob/{}/{}#L{}'.format(
            self.dataimport.revision.repo.url.rstrip('/'),
            self.dataimport.revision.hash,
            self.path,
            self.line,
        )


class DataFile(models.Model):
    id = models.UUIDField(primary_key=True, default=uuid.uuid4)
    name = models.CharField(max_length=100)
    size = models.PositiveIntegerField()
    hash = models.CharField(max_length=32)
    content_type = models.CharField(max_length=50)
    corpus = models.ForeignKey('documents.Corpus', on_delete=models.CASCADE, related_name='files')
    image = models.OneToOneField('images.Image', on_delete=models.SET_NULL, blank=True, null=True)

    class Meta:
        unique_together = (('corpus', 'hash'), )
        ordering = ['corpus', 'name']

    @property
    def staging_path(self):
        return os.path.join(settings.MEDIA_ROOT, str(self.id))


def repository_default_branches():
    '''
    This is needed to avoid re-using the same list instance
    as Repository.watched_branches default on new instances
    See Django warning postgres.E003
    '''
    return ['refs/heads/master']


class Repository(models.Model):
    id = models.UUIDField(primary_key=True, default=uuid.uuid4)
    url = models.URLField(unique=True)
    hook_token = models.CharField(max_length=250, unique=True)
    corpus = models.ForeignKey('documents.Corpus', on_delete=models.CASCADE, related_name='repos')
    credentials = models.ForeignKey(
        'users.OAuthCredentials', on_delete=models.CASCADE, related_name='repos', blank=True, null=True)
    watched_branches = ArrayField(models.CharField(max_length=50), default=repository_default_branches)
    provider_name = models.CharField(
        max_length=50,
        choices=[(p.__name__, p.display_name) for p in git_providers],
    )

    class Meta:
        verbose_name_plural = 'repositories'

    @property
    def provider_class(self):
        return get_provider(self.provider_name)

    @property
    def provider(self):
        return self.provider_class(credentials=self.credentials)

    @property
    def clone_dir(self):
        return os.path.join(settings.CELERY_WORKING_DIR, str(self.id))


class Revision(models.Model):
    id = models.UUIDField(primary_key=True, default=uuid.uuid4)
    repo = models.ForeignKey('dataimport.Repository', on_delete=models.CASCADE, related_name='revisions')
    hash = models.CharField(max_length=50)
    ref = models.CharField(max_length=50)
    message = models.TextField()
    author = models.CharField(max_length=50)

    class Meta:
        unique_together = (('repo', 'hash'), )

    @property
    def commit_url(self):
        return '{}/commit/{}'.format(self.repo.url.rstrip('/'), self.hash)

    def start_import(self):
        dataimport = self.dataimports.create(
            creator=self.repo.credentials.user,
            corpus=self.repo.corpus,
            mode=DataImportMode.Repository,
            state=DataImportState.Configured,
        )
        dataimport.start()
        return dataimport
