from abc import ABC, abstractmethod
from django.urls import reverse
from rest_framework.exceptions import NotAuthenticated, AuthenticationFailed, APIException, ValidationError
from gitlab import Gitlab, GitlabGetError, GitlabCreateError
from arkindex.documents.models import Corpus
import urllib.parse
import base64
import uuid


class GitProvider(ABC):

    display_name = None

    def __init__(self, credentials=None, url=None):
        if credentials:
            from arkindex.users.models import OAuthCredentials
            assert isinstance(credentials, OAuthCredentials)
            self.url = credentials.provider_url
        if url:
            self.url = url
        self.credentials = credentials

    @abstractmethod
    def list_repos(self, query=None):
        """
        List all repositories or filter with a search query.
        """

    @abstractmethod
    def create_repo(self, **kwargs):
        """
        Create a Repository instance from an external repository
        """

    @abstractmethod
    def download_archive(self, revision, path):
        """
        Download an archive for a given Revision instance.
        """

    @abstractmethod
    def get_or_create_latest_revision(self, repo):
        """
        Get a Revision instance for the last revision on the main branch of a given repository.
        """

    @abstractmethod
    def handle_webhook(self, repo, request):
        """
        Handle a webhook event on a given repository.
        """


class GitLabProvider(GitProvider):

    display_name = "GitLab"
    url = 'https://gitlab.com'

    def list_repos(self, query=None):
        if not self.credentials:
            raise NotAuthenticated()
        gl = Gitlab(self.url, oauth_token=self.credentials.token)
        return gl.projects.list(membership=True, search=query)

    def create_repo(self, id=None, corpus=None, request=None, **kwargs):
        assert isinstance(corpus, Corpus)
        if not self.credentials and request:
            raise NotAuthenticated()
        gl = Gitlab(self.url, oauth_token=self.credentials.token)
        try:
            project = gl.projects.get(int(id))
        except GitlabGetError as e:
            raise APIException("Error while fetching GitLab project: {}".format(str(e)))

        from arkindex.dataimport.models import Repository
        if Repository.objects.filter(url=project.web_url).exists():
            raise ValidationError("A repository with this URL already exists")

        repo = self.credentials.repos.create(
            corpus=corpus,
            url=project.web_url,
            watched_branches=['refs/heads/{}'.format(project.default_branch)],
            hook_token=str(base64.b64encode(uuid.uuid4().bytes)),
            provider_name=self.__class__.__name__,
        )

        try:
            project.hooks.create({
                'url': request.build_absolute_uri(
                    reverse('api:import-hook', kwargs={'pk': repo.id})
                ),
                'push_events': True,
                'token': repo.hook_token,
            })
        except GitlabCreateError as e:
            raise APIException("Error while creating GitLab hook: {}".format(str(e)))

        return repo

    def download_archive(self, revision, path):
        gl = Gitlab(self.url, oauth_token=revision.repo.credentials.token)
        try:
            project = gl.projects.get(urllib.parse.urlsplit(revision.repo.url).path.strip('/'))
        except GitlabGetError as e:
            raise APIException("Error while fetching GitLab project: {}".format(str(e)))

        with open(path, 'wb') as f:
            project.repository_archive(sha=revision.hash, streamed=True, action=f.write)

    def get_or_create_latest_revision(self, repo):
        gl = Gitlab(self.url, oauth_token=repo.credentials.token)
        try:
            project = gl.projects.get(urllib.parse.urlsplit(repo.url).path.strip('/'))
        except GitlabGetError as e:
            raise APIException("Error while fetching GitLab project: {}".format(str(e)))

        latest_commit = project.commits.list()[0]
        return repo.revisions.get_or_create(
            repo=repo,
            hash=latest_commit.id,
            defaults={
                'ref': latest_commit.refs()[0]['name'],
                'message': latest_commit.message,
                'author': latest_commit.author_name,
            },
        )

    def handle_webhook(self, repo, request):
        if 'HTTP_X_GITLAB_EVENT' not in request.META:
            raise ValidationError("Missing GitLab event type")
        if request.META['HTTP_X_GITLAB_EVENT'] != 'Push Hook':
            raise ValidationError("Unsupported GitLab event type")

        if 'HTTP_X_GITLAB_TOKEN' not in request.META:
            raise NotAuthenticated("Missing GitLab secret token")
        if request.META['HTTP_X_GITLAB_TOKEN'] != repo.hook_token:
            raise AuthenticationFailed("Invalid GitLab secret token")

        assert isinstance(request.data, dict)
        assert request.data['object_kind'] == 'push'

        if request.data['ref'] not in repo.watched_branches:
            return

        # Already took care of this event
        if repo.revisions.filter(
                ref=request.data['ref'],
                hash=request.data['checkout_sha']).exists():
            return

        rev = repo.revisions.create(
            hash=request.data['checkout_sha'],
            ref=request.data['ref'],
            message=request.data['commits'][0]['message'],
            author=request.data['commits'][0]['author']['name'],
        )
        rev.start_import()


git_providers = [
    GitLabProvider,
]
oauth_to_git = {
    "GitLabOAuthProvider": GitLabProvider,
}


def get_provider(name):
    return next(filter(lambda p: p.__name__ == name, git_providers), None)


def from_oauth(name):
    return oauth_to_git.get(name)
