from celery import shared_task
from celery.utils.log import get_task_logger
from celery.signals import task_postrun
from celery.states import EXCEPTION_STATES
from django.conf import settings
from django.db import transaction
from arkindex.project.celery import ReportingTask
from arkindex.documents.models import Element, ElementType, Page
from arkindex.documents.importer import import_page
from arkindex.documents.tei import TeiParser
from arkindex.images.models import ImageServer, ImageStatus
from arkindex.dataimport.models import DataImport, DataImportState, DataImportMode
from PIL import Image
import os
import glob
import logging
import shutil
import urllib.parse

root_logger = logging.getLogger(__name__)
logger = get_task_logger(__name__)


class TaskLoggingHandler(logging.Handler):

    def __init__(self, task, level=logging.WARNING):
        assert isinstance(task, ReportingTask)
        super().__init__(level=level)
        self.task = task

    def emit(self, record):
        self.task.report_message(record.getMessage(), level=record.levelno)


@shared_task(bind=True, base=ReportingTask)
def check_images(self, dataimport):
    assert isinstance(dataimport, DataImport)

    datafiles = dataimport.files.all()
    valid_files = []
    filecount = len(datafiles)

    for i, datafile in enumerate(datafiles):
        self.report_progress(i / filecount, "Checking image {} of {}".format(i + 1, filecount))

        try:
            img = Image.open(datafile.staging_path)
            assert max(img.size) >= 500, "Image {} is too small".format(datafile.name)
        except IOError:
            self.report_message("File {} is not a valid image".format(datafile.name), level=logging.WARNING)
            continue
        except AssertionError as e:
            self.report_message(str(e), level=logging.WARNING)
            continue

        valid_files.append(datafile)

    assert len(valid_files) > 0, "No valid JPEG images in selected files"
    return valid_files


@shared_task(bind=True, base=ReportingTask)
def import_images(self, valid_files, dataimport, server_id=settings.LOCAL_IMAGESERVER_ID):
    assert isinstance(dataimport, DataImport)
    self.report_progress(0, 'Pre import checks...')
    try:
        server = ImageServer.objects.get(id=server_id)
    except ImageServer.DoesNotExist:
        raise ValueError('Image server {} does not exist'.format(server_id))

    # Get volume
    try:
        volume = Element.objects.get(
            type=ElementType.Volume,
            corpus=dataimport.corpus,
            pk=dataimport.payload['volume_id'],
        )
    except Element.DoesNotExist:
        raise ValueError('Missing volume')

    if not os.path.exists(dataimport.iiif_path):
        os.makedirs(dataimport.iiif_path)

    datafiles = dataimport.files.all()
    pages = []
    for i, datafile in enumerate(datafiles, 1):
        self.report_progress(i / len(datafiles), 'Importing image {} of {}'.format(i, len(datafiles)))

        pillow_img = Image.open(datafile.staging_path)
        width, height = pillow_img.size

        # Non-JPEG image formats that should not be converted
        # Will default to .jpg if the image format is not in there
        # Formats are Pillow images formats,
        # see https://pillow.readthedocs.io/en/5.1.x/handbook/image-file-formats.html
        known_exts = {
            "JPEG2000": ".jp2",
            "PNG": ".png",
            "TIFF": ".tif",
        }
        ext = known_exts.get(pillow_img.format, ".jpg")
        newfilename = str(datafile.id) + ext

        if hasattr(datafile, 'image') and datafile.image is not None:
            self.report_message(
                "File {} is already linked to an image, skipping".format(datafile.name),
                level=logging.WARNING,
            )
            continue

        destination_path = os.path.join(dataimport.iiif_path, newfilename)
        pillow_img.save(destination_path)

        img, _ = server.images.get_or_create(
            path=urllib.parse.urljoin(dataimport.folder_name + '/', newfilename),
            defaults={
                'width': width,
                'height': height,
                'status': ImageStatus.Checked,
                'datafile': datafile,
            }
        )

        page = import_page(volume, img, volume.name)
        pages.append((page.id, img.get_thumbnail_url(max_width=500)))

    assert len(pages) > 0, 'No imported files'
    self.report_message("Imported files into {}".format(volume))
    return {
        'volume': str(volume.id),
        'pages': pages,
    }


@shared_task(bind=True, base=ReportingTask)
def save_ml_results(self, results, **kwargs):
    '''
    Save ML results on images into our DB:
    * Classification
    * Text extraction
    '''
    assert isinstance(results, dict)

    pages = Page.objects.filter(pk__in=list(results.keys()))
    for page in pages:
        result = results.get(page.id)
        if result is None:
            continue

        page.classification = result['classification']
        page.text = result['text']
        page.save()

        self.report_message("Updated ML results for {}".format(page))

    return list(map(str, results.keys()))


@shared_task(bind=True, base=ReportingTask)
def download_repo(self, dataimport):
    assert isinstance(dataimport, DataImport)
    assert dataimport.mode == DataImportMode.Repository
    assert dataimport.revision is not None

    repo_dir = dataimport.revision.repo.clone_dir
    if os.path.exists(repo_dir):
        shutil.rmtree(repo_dir)
    archive_path = "{}.tar.gz".format(repo_dir)
    commit_hash = dataimport.revision.hash

    if dataimport.revision.repo.provider_class is None:
        raise ValueError("No repository provider found for {}".format(dataimport.revision.repo.url))

    self.report_progress(0, "Downloading repository archive at {}...".format(commit_hash))
    dataimport.revision.repo.provider.download_archive(dataimport.revision, archive_path)

    self.report_progress(0.5, "Extracting...")
    shutil.unpack_archive(archive_path, repo_dir)
    os.remove(archive_path)


@shared_task(bind=True, base=ReportingTask)
def import_repo(self, dataimport):
    handler = TaskLoggingHandler(self)
    root_logger.addHandler(handler)

    self.report_progress(0, "Finding XML files...")
    xml_files = glob.glob(os.path.join(dataimport.revision.repo.clone_dir, '**/*.xml'), recursive=True)

    if len(xml_files) < 1:
        self.report_message("No XML files found.", level=logging.WARNING)

    for i, xml_file in enumerate(xml_files, 1):
        filename = os.path.basename(xml_file)
        self.report_progress(i / len(xml_files), 'Importing file {} of {}: {}'.format(i, len(xml_files), filename))

        try:
            parser = TeiParser(xml_file)
            parser.check()
            matches = parser.match_database(dataimport.corpus)
            dataimport.save_failures(parser.match_errors)

            for db_elt, tei_elt in matches:
                with transaction.atomic():
                    tei_elt.save(db_elt, dataimport.revision)

        except Exception as e:
            self.report_message(
                "Importing of {} failed: {}".format(filename, str(e)), level=logging.WARNING)

    root_logger.removeHandler(handler)


@shared_task(bind=True, base=ReportingTask)
def cleanup_repo(self, dataimport):
    shutil.rmtree(dataimport.revision.repo.clone_dir)


@task_postrun.connect
def dataimport_postrun(task_id, task, state, args=(), **kwargs):
    '''
    Update DataImport state after task exit
    '''

    # Only for dataimport or machine learning tasks
    if not any(map(task.name.startswith, ('arkindex.dataimport.tasks', 'arkindex_ml.tasks'))):
        return

    # Look for dataimport in args
    imps = [a for a in args if isinstance(a, DataImport)]
    if len(imps) == 1:
        dataimport = imps[0]
        dataimport.refresh_from_db()  # avoid inconsistency

    elif 'kwargs' in kwargs and 'dataimport_id' in kwargs['kwargs']:
        # Try to load a dataimport from its id
        # Needed to communicate cleanly with worker-ml (no django instances)
        try:
            dataimport = DataImport.objects.get(pk=kwargs['kwargs']['dataimport_id'])
        except DataImport.DoesNotExists:
            raise Exception('No dataimport found in args')

    else:
        raise Exception('No args on dataimport task.')

    assert isinstance(dataimport, DataImport), \
        'DataImport not found as first arg of task {}'.format(task)

    # Manage error state
    if any(t.failed() for t in dataimport.tasks) or state in EXCEPTION_STATES:
        dataimport.state = DataImportState.Error

        # Report exception message
        if 'retval' in kwargs:
            task.report_message(str(kwargs['retval']), logging.ERROR)

    # Manage completion when all tasks are completed
    elif sum(t.successful() for t in dataimport.tasks) == dataimport.task_count:
        dataimport.state = DataImportState.Done

    dataimport.save()
