from django.contrib import admin
from django.conf.urls import url
from django.urls import reverse
from django.utils.html import format_html
from arkindex.documents.models import Corpus, Page, Element, ElementType, Act, Transcription, MetaData
from arkindex.documents.views import DumpActs
from enumfields.admin import EnumFieldListFilter


class CorpusAdmin(admin.ModelAdmin):
    list_display = ('id', 'name', 'public', )


class PageAdmin(admin.ModelAdmin):
    list_display = ('id', 'name', 'page_type', 'nb', 'direction', )
    list_filter = [('page_type', EnumFieldListFilter)]
    fields = ('id', 'name', 'folio', 'page_type', 'nb', 'direction', 'classification', 'text')
    readonly_fields = ('id', )


class MetaDataInline(admin.TabularInline):
    model = MetaData


class ElementAdmin(admin.ModelAdmin):
    list_display = ('id', 'name', 'type', 'corpus', 'element_actions')
    list_filter = [('type', EnumFieldListFilter), 'corpus']
    fields = ('id', 'type', 'name', 'corpus')
    readonly_fields = ('id', 'element_actions')
    inlines = (MetaDataInline, )

    def get_urls(self):
        urls = super().get_urls()
        custom_urls = [
            url(
                r'^(?P<pk>.+)/acts/$',
                self.admin_site.admin_view(DumpActs.as_view()),
                name='dump-acts',
            ),
        ]
        return custom_urls + urls

    def element_actions(self, obj):
        return format_html(
            '<a class="button" href="{}">Dump acts</a>',
            reverse('admin:dump-acts', args=[obj.pk]),
        ) if obj.type == ElementType.Volume else ''
    element_actions.short_description = 'Actions'
    element_actions.allow_tags = True


class ActAdmin(admin.ModelAdmin):
    list_display = ('id', 'name')
    fields = ('id', 'name', 'folio', 'number')
    readonly_fields = ('id', )
    inlines = (MetaDataInline, )


class TranscriptionAdmin(admin.ModelAdmin):
    list_display = ('id', 'text', 'score', 'element', )
    list_filter = [('type', EnumFieldListFilter), ]
    fields = ('id', 'text', 'score', 'element', 'zone', )
    readonly_fields = ('id', )
    raw_id_fields = ('element', 'zone', )


admin.site.register(Corpus, CorpusAdmin)
admin.site.register(Page, PageAdmin)
admin.site.register(Element, ElementAdmin)
admin.site.register(Act, ActAdmin)
admin.site.register(Transcription, TranscriptionAdmin)
