from django.conf import settings
from django.utils.decorators import method_decorator
from django.views.decorators.cache import cache_page
from rest_framework.generics import RetrieveAPIView
from arkindex.documents.models import Element, ElementType, Page, Act, Transcription, Corpus
from arkindex.documents.serializers.iiif import \
    VolumeManifestSerializer, ActManifestSerializer, \
    PageAnnotationListSerializer, PageActAnnotationListSerializer, \
    SurfaceAnnotationListSerializer, TranscriptionSearchAnnotationListSerializer
from arkindex.documents.search import search_transcriptions_filter_post
from arkindex.project.elastic import ESQuerySet


class VolumeManifest(RetrieveAPIView):
    """
    Get a IIIF manifest for a specific volume
    """

    serializer_class = VolumeManifestSerializer

    def get_queryset(self):
        return Element.objects.filter(
            type=ElementType.Volume,
            corpus__in=Corpus.objects.readable(self.request.user),
        )

    @method_decorator(cache_page(3600))
    def get(self, *args, **kwargs):
        return super().get(*args, **kwargs)


class ActManifest(RetrieveAPIView):
    """
    Get a IIIF manfiest for a specific act
    """

    serializer_class = ActManifestSerializer
    queryset = Act.objects.all()

    def get_queryset(self):
        return Act.objects.filter(corpus__in=Corpus.objects.readable(self.request.user))

    @method_decorator(cache_page(3600))
    def get(self, *args, **kwargs):
        return super().get(*args, **kwargs)


class PageAnnotationList(RetrieveAPIView):
    """
    Get a IIIF annotation list for transcriptions of a specific page
    """

    serializer_class = PageAnnotationListSerializer

    def get_queryset(self):
        return Page.objects.filter(corpus__in=Corpus.objects.readable(self.request.user))

    @method_decorator(cache_page(3600))
    def get(self, *args, **kwargs):
        return super().get(*args, **kwargs)


class PageActAnnotationList(RetrieveAPIView):
    """
    Get a IIIF annotation list for acts of a specific page
    """

    serializer_class = PageActAnnotationListSerializer

    def get_queryset(self):
        return Page.objects.filter(corpus__in=Corpus.objects.readable(self.request.user))

    @method_decorator(cache_page(3600))
    def get(self, *args, **kwargs):
        return super().get(*args, **kwargs)


class SurfaceAnnotationList(RetrieveAPIView):
    """
    Get a IIIF annotation list for a single surface
    """
    serializer_class = SurfaceAnnotationListSerializer

    def get_queryset(self):
        return Element.objects.filter(
            type=ElementType.Surface,
            corpus__in=Corpus.objects.readable(self.request.user),
        )

    @method_decorator(cache_page(3600))
    def get(self, *args, **kwargs):
        return super().get(*args, **kwargs)


class TranscriptionSearchAnnotationList(RetrieveAPIView):
    """
    Search for transcriptions inside an element and get a IIIF annotation list
    """

    serializer_class = TranscriptionSearchAnnotationListSerializer

    def get_object(self):
        elt = Element.objects.get(
            corpus__in=Corpus.objects.readable(self.request.user),
            id=self.kwargs['pk'],
        )
        return ESQuerySet(
            query=ESQuerySet.make_query(
                'elastic/search_transcriptions.json',
                ctx={
                    'query': self.request.query_params.get('q'),
                    'type': 'word',
                    'corpus_id': elt.corpus_id,
                    'inner_hits_size': settings.ES_INNER_RESULTS_LIMIT,
                },
            ),
            es_index=settings.ES_INDEX_TRANSCRIPTIONS,
            es_type=Transcription.INDEX_TYPE,
            post_process=search_transcriptions_filter_post,
            post_process_args=[elt.id]
        )
