#!/usr/bin/env python3

from django.core.management.base import BaseCommand, CommandError
from django.db import transaction
from arkindex.documents.models import Corpus
from arkindex.documents.tei import TeiParser
import logging

logging.basicConfig(
    format='[%(levelname)s] %(message)s',
    level=logging.INFO
)
logger = logging.getLogger(__name__)


class Command(BaseCommand):
    help = "Import TEI metadata"

    def add_arguments(self, parser):
        parser.add_argument(
            'path',
            type=open,
            help='Path to TEI file'
        )
        parser.add_argument(
            '--corpus',
            type=str,
            help='Slug of corpus to import metadata into',
            required=True,
        )
        parser.add_argument(
            '--dry-run',
            action='store_true',
            default=False,
            help='Do not hit the Database'
        )

    def handle(self, *args, **options):
        # Handle verbosity level
        verbosity = int(options['verbosity'])
        root_logger = logging.getLogger('')
        if verbosity > 1:
            root_logger.setLevel(logging.DEBUG)

        # Load corpus
        try:
            corpus = Corpus.objects.get(id=options['corpus'])
        except Corpus.DoesNotExist:
            raise CommandError('Invalid corpus')

        # Initial check
        parser = TeiParser(options['path'].name)
        parser.check()

        # Match items
        matches = parser.match_database(corpus)
        logger.info('Found {} DB matches'.format(len(matches)))

        if options['dry_run']:

            # Display metadata
            for db_elt, tei_elt in matches:
                print(db_elt, tei_elt.build_metadata())

        # Save in database
        else:
            for db_elt, tei_elt in matches:
                with transaction.atomic():
                    # Remove old metadatas
                    db_elt.metadatas.all().delete()

                    # Create new ones
                    for name, meta in tei_elt.build_metadata().items():
                        if not meta[1]:
                            continue
                        db_elt.metadatas.create(
                            name=name,
                            type=meta[0],
                            value=meta[1],
                        )
