from arkindex.documents.models import Transcription, Act, Element
from itertools import chain
import uuid


def search_transcriptions_post(data):
    '''
    Search pages with a transcription
    List pages ordered by document
    '''
    results = data['hits']['hits']
    transcription_ids = [r['_id'] for r in results]
    ts = Transcription.objects \
                      .filter(id__in=transcription_ids) \
                      .order_by('-score') \
                      .prefetch_related('zone__image__server', 'element')
    # Recompute in case some transcriptions are not found in Postgres
    transcription_ids = [t.element_id for t in ts]
    all_parent_paths = Element.objects.get_ascendings_paths(*transcription_ids)
    for trans in ts:
        trans.parent_paths = all_parent_paths.get(trans.element_id, [])
    return ts


def search_acts_post(data):
    """
    Search acts containing query
    Returns (acts list, total number of acts)
    """

    results = data['hits']['hits']
    act_ids = [uuid.UUID(r['_id']) for r in results]
    act_scores = {uuid.UUID(r['_id']): r['_score'] for r in results}
    if not act_ids:
        return
    tr_ids = [
        hit['_source']['id']
        for result in results
        for hit in result['inner_hits']['transcriptions']['hits']['hits']
    ]
    tr_totals = {
        uuid.UUID(result['_id']): result['inner_hits']['transcriptions']['hits']['total']
        for result in results
    }

    transcriptions = {
        t.id: t
        for t in Transcription.objects.filter(id__in=tr_ids).prefetch_related('zone__image__server')
    }

    acts = Act.objects.filter(id__in=act_ids).prefetch_related('corpus')
    acts_tr_ids = {
        uuid.UUID(result['_id']): [
            uuid.UUID(hit['_source']['id'])
            for hit in result['inner_hits']['transcriptions']['hits']['hits']
        ] for result in results
    }

    all_paths = Element.objects.get_ascendings_paths(*act_ids)

    for act in acts:
        act.transcriptions_results = [transcriptions[tid] for tid in acts_tr_ids[act.id]]
        act.total_transcriptions = tr_totals[act.id]
        act.parent_paths = all_paths.get(act.id, [])

    return sorted(acts, key=lambda a: act_scores[a.id], reverse=True)


def search_transcriptions_filter_post(data, element_id):
    if not isinstance(element_id, uuid.UUID):
        element_id = uuid.UUID(element_id)
    return filter(
        lambda t: element_id in [p.id for p in chain(*t.parent_paths)],
        search_transcriptions_post(data)
    )
