from arkindex.project.tests import FixtureAPITestCase
from arkindex.documents.models import Transcription, Act, Element, Corpus
from django.urls import reverse
from django.contrib.auth.models import AnonymousUser
from rest_framework import status
from unittest.mock import patch


class TestSearchAPI(FixtureAPITestCase):

    @classmethod
    def setUpTestData(cls):
        cls.es_mock = patch('arkindex.project.elastic.Elasticsearch').start()

    def setUp(self):
        self.es_mock.reset_mock()
        self.es_mock().reset_mock()

    def build_es_response(self, hits):
        return {
            "hits": {
                "total": len(hits),
                "max_score": None,
                "hits": hits
            },
            "_shards": {
                "total": 5,
                "failed": 0,
                "skipped": 0,
                "successful": 5
            },
            "took": 42,
            "timed_out": False
        }

    def make_transcription_hit(self, ts):
        return {
            "_id": str(ts.id.hex),
            "_score": None,
            "sort": [
                ts.score
            ],
            "_index": "transcriptions",
            "_type": Transcription.INDEX_TYPE,
            "_source": ts.build_search_index()
        }

    def make_nested_transcription_hit(self, ts):
        return {
            "_source": {
                "text": ts.text,
                "id": str(ts.id),
                "score": ts.score,
            },
            "_score": ts.score,
            "_nested": {
                "field": "transcriptions",
                "offset": 1337
            }
        }

    def make_act_hit(self, act, ts, score=1.0):
        return {
            "_score": score,
            "_type": Act.INDEX_TYPE,
            "_id": str(act.id.hex),
            "_index": "acts",
            "inner_hits": {
                "transcriptions": {
                    "hits": {
                        "total": len(ts),
                        "hits": list(map(self.make_nested_transcription_hit, ts)),
                        "max_score": 1337,
                    }
                }
            }
        }

    def test_transcription_search(self):
        expected = Transcription.objects.filter(text="PARIS")

        self.es_mock().count.return_value = {'count': len(expected)}
        self.es_mock().search.return_value = self.build_es_response(
            list(map(self.make_transcription_hit, expected)),
        )

        response = self.client.get(reverse('api:transcription-search'), {'q': "paris"})
        self.assertEqual(response.status_code, status.HTTP_200_OK)

        results = response.json()["results"]

        self.assertCountEqual(
            [r['id'] for r in results],
            map(str, expected.values_list('id', flat=True)),
        )

        args, kwargs = self.es_mock().search.call_args
        self.assertTupleEqual(args, ())
        self.assertCountEqual(kwargs.keys(), ['body', 'index', 'doc_type'])
        self.assertEqual(kwargs['index'], 'transcriptions')
        self.assertEqual(kwargs['doc_type'], Transcription.INDEX_TYPE)

        self.assertCountEqual(kwargs['body'].keys(), ['_source', 'from', 'size', 'query', 'sort', 'aggs'])
        self.assertEqual(kwargs['body']['_source'], True)
        self.assertEqual(kwargs['body']['from'], 0)
        self.assertEqual(kwargs['body']['size'], len(expected))

        self.assertIsInstance(kwargs['body']['query']['bool']['must'], list)
        self.assertTrue(all(len(cond.keys()) == 1 for cond in kwargs['body']['query']['bool']['must']))
        conditions = {
            list(cond.keys())[0]: list(cond.values())[0]
            for cond in kwargs['body']['query']['bool']['must']
        }

        self.assertCountEqual(conditions.keys(), ['match', 'terms', 'range'])
        self.assertEqual(conditions['match']['text'], 'paris')
        self.assertEqual(conditions['range']['score']['gte'], 0.5)
        self.assertCountEqual(
            conditions['terms']['corpus.keyword'],
            map(str, Corpus.objects.readable(AnonymousUser()).values_list('id', flat=True)),
        )

    def test_act_search(self):
        act = Act.objects.get(number="1")
        ts = Transcription.objects.filter(text__in=["PARIS", "ROY"], zone__image__path='img1')

        self.es_mock().count.return_value = {'count': 1}
        self.es_mock().search.return_value = self.build_es_response(
            [self.make_act_hit(act, ts), ],
        )

        response = self.client.get(reverse('api:act-search'), {'q': "paris roy"})
        self.assertEqual(response.status_code, status.HTTP_200_OK)

        results = response.json()["results"]
        self.assertEqual(len(results), 1)
        result = results[0]

        self.assertEqual(result['id'], str(act.id))
        self.assertCountEqual(
            [t['id'] for t in result['transcriptions']],
            map(str, ts.values_list('id', flat=True)),
        )
        self.assertEqual(result['total_transcriptions'], len(ts))

        args, kwargs = self.es_mock().search.call_args
        self.assertTupleEqual(args, ())
        self.assertCountEqual(kwargs.keys(), ['body', 'index', 'doc_type'])
        self.assertEqual(kwargs['index'], 'acts')
        self.assertEqual(kwargs['doc_type'], Act.INDEX_TYPE)

        self.assertCountEqual(kwargs['body'].keys(), ['_source', 'from', 'size', 'query', 'sort', 'aggs'])
        self.assertEqual(kwargs['body']['_source'], False)
        self.assertEqual(kwargs['body']['from'], 0)
        self.assertEqual(kwargs['body']['size'], 1)

        self.assertEqual(kwargs['body']['query']['nested']['score_mode'], 'sum')
        self.assertEqual(kwargs['body']['query']['nested']['path'], 'transcriptions')
        self.assertIn('inner_hits', kwargs['body']['query']['nested'])

        function_score = kwargs['body']['query']['nested']['query']['function_score']
        self.assertListEqual(function_score['functions'], [
            {
                "field_value_factor": {
                    "field": "transcriptions.score",
                }
            }
        ])
        self.assertIsInstance(function_score['query']['bool']['must'], list)

        self.assertTrue(all(len(cond.keys()) == 1 for cond in function_score['query']['bool']['must']))
        conditions = {
            list(cond.keys())[0]: list(cond.values())[0]
            for cond in function_score['query']['bool']['must']
        }
        self.assertCountEqual(conditions.keys(), ['match', 'terms', 'range'])
        self.assertEqual(conditions['match']['transcriptions.text'], 'paris roy')
        self.assertEqual(conditions['range']['transcriptions.score']['gte'], 0.5)
        self.assertCountEqual(
            conditions['terms']['transcriptions.corpus'],
            map(str, Corpus.objects.readable(AnonymousUser()).values_list('id', flat=True)),
        )

    def test_iiif_transcription_search(self):
        # Filter to only get transcriptions from volume 1
        unfiltered = Transcription.objects.filter(text="PARIS")
        expected = Transcription.objects.filter(text="PARIS", zone__image__path__in=['img1', 'img2', 'img3'])
        vol = Element.objects.get(name='Volume 1')

        self.es_mock().count.return_value = {'count': len(unfiltered)}
        self.es_mock().search.return_value = self.build_es_response(
            list(map(self.make_transcription_hit, unfiltered))
        )

        response = self.client.get(reverse('api:ts-search-manifest', kwargs={'pk': str(vol.id)}), {'q': 'paris'})
        self.assertEqual(response.status_code, status.HTTP_200_OK)
        data = response.json()

        self.assertEqual(data['@context'], "http://iiif.io/api/search/0/context.json")
        self.assertEqual(data['@type'], 'sc:AnnotationList')
        self.assertEqual(data['startIndex'], 0)
        self.assertEqual(data['within']['@type'], 'sc:Layer')
        self.assertEqual(data['within']['total'], len(expected))

        hits = data['hits']
        self.assertTrue(all(hit['@type'] == 'search:Hit' for hit in hits))
        self.assertTrue(all(hit['match'] == 'PARIS' for hit in hits))

        resources = data['resources']
        self.assertTrue(all(res['@type'] == 'oa:Annotation' for res in resources))
        self.assertTrue(all(res['motivation'] == 'sc:painting' for res in resources))
        self.assertTrue(all(res['resource']['@type'] == 'cnt:ContentAsText' for res in resources))
        self.assertTrue(all(res['resource']['format'] == 'text/plain' for res in resources))
        self.assertTrue(all(res['resource']['chars'] == 'PARIS' for res in resources))

        args, kwargs = self.es_mock().search.call_args
        self.assertTupleEqual(args, ())
        self.assertCountEqual(kwargs.keys(), ['body', 'index', 'doc_type'])
        self.assertEqual(kwargs['index'], 'transcriptions')
        self.assertEqual(kwargs['doc_type'], Transcription.INDEX_TYPE)

        self.assertCountEqual(kwargs['body'].keys(), ['_source', 'from', 'size', 'query', 'sort', 'aggs'])
        self.assertEqual(kwargs['body']['_source'], True)
        self.assertEqual(kwargs['body']['from'], 0)
        self.assertEqual(kwargs['body']['size'], len(expected))

        self.assertIsInstance(kwargs['body']['query']['bool']['must'], list)
        self.assertTrue(all(len(cond.keys()) == 1 for cond in kwargs['body']['query']['bool']['must']))
        conditions = kwargs['body']['query']['bool']['must']

        text_match = next(c['match']['text'] for c in conditions if 'match' in c and 'text' in c['match'])
        corpus_match = next(c['match']['corpus'] for c in conditions if 'match' in c and 'corpus' in c['match'])
        self.assertEqual(text_match, 'paris')
        self.assertEqual(corpus_match, str(vol.corpus_id))
        self.assertEqual(next(c['range']['score']['gte'] for c in conditions if 'range' in c), 0.5)
