from arkindex.project.tests import FixtureTestCase
from arkindex.documents.models import Element, Page, PageDirection, PageComplement
from arkindex.documents.surface_link import parse_folios, ParsedFolio


class TestSurfaceLinkerFunctions(FixtureTestCase):
    """Test SurfaceLinker helper functions."""

    def setUp(self):
        vol = Element.objects.get(name="Volume 1")
        self.p1r = Page.objects.get(name="Volume 1, page 1r")
        self.p1v = Page.objects.get(name="Volume 1, page 1v")
        self.p2r = Page.objects.get(name="Volume 1, page 2r")
        self.p2bisr = Page.objects.create(
            corpus=self.corpus,
            name="2bisr",
            folio="2bisr",
            nb=2,
            direction=PageDirection.Recto,
            complement=PageComplement.Bis,
        )
        self.p2v = Page.objects.create(corpus=self.corpus, name="2v", folio="2v", nb=2, direction=PageDirection.Verso)
        for page in (self.p2bisr, self.p2v):
            page.add_parent(vol)
        self.pages = list(Page.objects.get_descending(vol.id))

    def test_page_order(self):
        self.assertTrue(self.p1r < self.p1v)
        self.assertTrue(self.p1v < self.p2r)
        self.assertTrue(self.p1r < self.p2r)
        self.assertTrue(self.p1r < self.p2v)
        self.assertTrue(self.p2r < self.p2v)
        self.assertTrue(self.p1r < self.p2bisr)
        self.assertTrue(self.p1v < self.p2bisr)
        self.assertTrue(self.p2r < self.p2bisr)
        self.assertTrue(self.p2bisr < self.p2v)
        self.assertTrue(self.p2v == self.p2v)

    def test_parse_folios_no_check(self):
        "Test parse_folios without a page queryset"

        folios = parse_folios('14r')
        self.assertEqual(len(folios), 1)
        self.assertIsInstance(folios[0], ParsedFolio)
        self.assertEqual(folios[0].number, 14)
        self.assertEqual(folios[0].complement, None)
        self.assertEqual(folios[0].direction, PageDirection.Recto)

        folios = parse_folios('17bisv-019quat')
        self.assertEqual(len(folios), 2)
        self.assertIsInstance(folios[0], ParsedFolio)
        self.assertEqual(folios[0].number, 17)
        self.assertEqual(folios[0].complement, PageComplement.Bis)
        self.assertEqual(folios[0].direction, PageDirection.Verso)
        self.assertIsInstance(folios[1], ParsedFolio)
        self.assertEqual(folios[1].number, 19)
        self.assertEqual(folios[1].complement, PageComplement.Quat)
        self.assertEqual(folios[1].direction, PageDirection.Recto)

        with self.assertRaisesMessage(ValueError, 'Improperly formatted folio string'):
            parse_folios('this is not a folio')

    def test_parse_folios_ignore_check(self):
        "Test parse_folios with a page queryset: ignore if there is only one folio"

        folios = parse_folios('14r', pages=self.pages)
        self.assertEqual(len(folios), 1)
        self.assertIsInstance(folios[0], ParsedFolio)
        self.assertEqual(folios[0].number, 14)
        self.assertEqual(folios[0].complement, None)
        self.assertEqual(folios[0].direction, PageDirection.Recto)

    def test_parse_folios_check_success(self):
        "Test parse_folios with a page queryset and a valid folio"

        folios = parse_folios('2r-2v', pages=self.pages)
        self.assertEqual(len(folios), 3)
        self.assertIsInstance(folios[0], ParsedFolio)
        self.assertEqual(folios[0].number, 2)
        self.assertEqual(folios[0].complement, None)
        self.assertEqual(folios[0].direction, PageDirection.Recto)
        self.assertIsInstance(folios[1], ParsedFolio)
        self.assertEqual(folios[1].number, 2)
        self.assertEqual(folios[1].complement, PageComplement.Bis)
        self.assertEqual(folios[1].direction, PageDirection.Recto)
        self.assertIsInstance(folios[2], ParsedFolio)
        self.assertEqual(folios[2].number, 2)
        self.assertEqual(folios[2].complement, None)
        self.assertEqual(folios[2].direction, PageDirection.Verso)

    def test_parse_folios_wrong_order(self):
        "Test parse_folios with a page queryset and a folio in a wrong order"

        with self.assertRaisesMessage(ValueError, 'Folio seems to be in reverse order'):
            parse_folios('1v-1r', pages=self.pages)
