from arkindex.project.tests import FixtureTestCase
from arkindex.project.polygon import Polygon
from arkindex.documents.models import Transcription, TranscriptionType
from arkindex.images.models import Image
from arkindex.images.importer import bulk_transcriptions


class TestBulkTranscriptions(FixtureTestCase):
    """Tests for bulk transcription and zone importing"""

    def setUp(self):
        # Create a page and an image
        self.img = Image.objects.get(path='img6', server=self.imgsrv)
        pagezone = self.img.zones.first()
        self.page = pagezone.elements.first()

    def test_bulk_transcriptions(self):
        items = [
            {
                'x': 0,
                'y': 0,
                'width': 100,
                'height': 100,
                'text': 'test 1',
                'score': 0.1,
            },
            {
                'x': 20,
                'y': 20,
                'width': 100,
                'height': 100,
                'text': 'test 2',
                'score': 0.2,
            },
        ]
        bulk_transcriptions(self.img, self.page, items)
        out = Transcription.objects.filter(zone__image=self.img).order_by('score')

        self.assertEqual(len(out), 2)
        self.assertIsInstance(out[0], Transcription)
        self.assertIsInstance(out[1], Transcription)

        self.assertEqual(out[0].text, 'test 1')
        self.assertEqual(out[0].score, 0.1)

        self.assertEqual(out[1].text, 'test 2')
        self.assertEqual(out[1].score, 0.2)

        self.assertIsNotNone(out[0].zone)
        self.assertIsNotNone(out[1].zone)

        self.assertEqual(out[0].zone.polygon, Polygon.from_coords(0, 0, 100, 100))
        self.assertEqual(out[1].zone.polygon, Polygon.from_coords(20, 20, 100, 100))

        # Check path
        children = self.page.transcriptions.all()
        self.assertEqual(children.count(), 2)
        ids = children.values_list('id', flat=True)
        self.assertIn(out[0].id, ids)
        self.assertIn(out[1].id, ids)

    def test_bulk_transcriptions_unique(self):
        """Check bulk_transcriptions does not import the same transcriptions twice"""
        items = [
            {
                'x': 0,
                'y': 0,
                'width': 100,
                'height': 100,
                'text': 'test 1',
                'score': 0.1,
            },
            {
                'x': 20,
                'y': 20,
                'width': 100,
                'height': 100,
                'text': 'test 2',
                'score': 0.2,
            },
        ]
        out = bulk_transcriptions(self.img, self.page, items)
        self.assertEqual(len(out), 2)
        out = bulk_transcriptions(self.img, self.page, items)
        self.assertEqual(len(out), 0)

    def test_bulk_transcriptions_polygons(self):
        items = [
            {
                'polygon': [[0, 0], [0, 200], [200, 200], [200, 0], [250, 50], [0, 0]],
                'text': 'test 1',
                'score': 0.1,
                'type': TranscriptionType.Word,
            },
            {
                'polygon': [[0, 0], [100, 0], [100, 100], [0, 100], [0, 0]],
                'text': 'test 2',
                'score': 0.2,
                'type': TranscriptionType.Line,
            },
        ]
        out = bulk_transcriptions(self.img, self.page, items)
        self.assertEqual(len(out), 2)

        trs = Transcription.objects.filter(zone__image=self.img).order_by('score')

        self.assertEqual(trs.count(), 2)
        self.assertIsInstance(trs[0], Transcription)
        self.assertIsInstance(trs[1], Transcription)

        self.assertEqual(trs[0].type, TranscriptionType.Word)
        self.assertEqual(trs[0].text, 'test 1')
        self.assertEqual(trs[0].score, 0.1)

        self.assertEqual(trs[1].type, TranscriptionType.Line)
        self.assertEqual(trs[1].text, 'test 2')
        self.assertEqual(trs[1].score, 0.2)

        self.assertIsNotNone(trs[0].zone)
        self.assertIsNotNone(trs[1].zone)

        self.assertEqual(trs[0].zone.polygon, Polygon([[0, 0], [0, 200], [200, 200], [200, 0], [250, 50], [0, 0]]))
        self.assertEqual(trs[1].zone.polygon, Polygon.from_coords(0, 0, 100, 100))

        # should not recreate
        out = bulk_transcriptions(self.img, self.page, items)
        self.assertEqual(len(out), 0)
