from __future__ import absolute_import, unicode_literals
import os
import time
import logging
from celery import Celery, Task
from celery.utils.log import get_task_logger
from celery.backends.redis import RedisBackend
from kombu.utils.encoding import bytes_to_str
from django.conf import settings

# set the default Django settings module for the 'celery' program.
os.environ.setdefault('DJANGO_SETTINGS_MODULE', 'arkindex.project.settings')

app = Celery('arkindex')

app.loader.override_backends['redis'] = 'arkindex.project.celery:ExtendedRedisBackend'

# Using a string here means the worker doesn't have to serialize
# the configuration object to child processes.
# - namespace='CELERY' means all celery-related configuration keys
#   should have a `CELERY_` prefix.
app.config_from_object('django.conf:settings', namespace='CELERY')

# celery_once uses conf.ONCE, but config_from_object imports everything
# in lowercase. This properly sets conf.ONCE.
app.conf.ONCE = settings.CELERY_ONCE

# Load task modules from all registered Django app configs.
app.autodiscover_tasks()


logger = get_task_logger(__name__)


class ExtendedRedisBackend(RedisBackend):

    def add_message(self, task_id, message):
        '''
        Add a message into the backend, on a separate redis key
        Used by our code only
        '''
        key = self.get_key_for_task(task_id)
        existing = self.get(key)
        meta = existing and self.decode(existing) or {}
        if 'messages' not in meta:
            meta['messages'] = []
        meta['messages'].append(message)
        self.set(key, self.encode(meta))

    def _store_result(self, task_id, result, state,
                      traceback=None, request=None, **kwargs):
        key = self.get_key_for_task(task_id)

        # Load existing messages
        existing = self.get(key)
        messages = existing and self.decode(existing).get('messages') or []

        meta = {
            'status': state,
            'result': result,
            'traceback': traceback,
            'children': self.current_task_children(request),
            'task_id': bytes_to_str(task_id),

            # Added for our usage
            'messages': messages,
            'task_name': request and str(request.task),
        }
        if request and getattr(request, 'group', None):
            meta['group_id'] = request.group
        self.set(key, self.encode(meta))
        return result


class ReportingTask(Task):

    def report_progress(self, progress, message=None):
        assert 0.0 <= progress <= 1.0
        self.update_state(state='PROGRESS', meta={'progress': progress})

        # Report message as info
        if not isinstance(message, str):
            message = '{:.0%}'.format(progress)
        self.report_message(message)

    def report_message(self, message, level=logging.INFO):
        assert isinstance(message, str)
        logger.log(msg=message, level=level)
        self.backend.add_message(
            self.request.id,
            {
                'level': level,
                'message': message,
                'time': time.time(),
            }
        )


@app.task(bind=True)
def debug_task(self):
    print('Request: {0!r}'.format(self.request))
