from django.contrib.postgres import fields
from django.db.models import Transform


class ArrayField(fields.ArrayField):
    """
    An enhanced PostgreSQL ArrayField, adding a `last` transform
    """

    def get_transform(self, name):
        """
        Get a Transform subclass corresponding to a name
        """
        transform = super().get_transform(name)
        if transform:
            return transform
        if name == 'last':
            return LastItemTransformFactory(self.base_field)


class LastItemTransform(Transform):
    """
    Get the last item of an ArrayField
    Requires array to respect PostgreSQL's indexing convention (start at 1)
    """

    def __init__(self, base_field, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.base_field = base_field

    def as_sql(self, compiler, connection):
        lhs, params = compiler.compile(self.lhs)
        return '%s[array_length(%s, 1)]' % (lhs, lhs), params

    @property
    def output_field(self):
        return self.base_field


class LastItemTransformFactory(object):
    """
    Create a LastItemTransform with a given base field
    """

    def __init__(self, base_field):
        self.base_field = base_field

    def __call__(self, *args, **kwargs):
        return LastItemTransform(self.base_field, *args, **kwargs)
