from django.conf import settings
from django.shortcuts import get_object_or_404
from rest_framework.exceptions import PermissionDenied, ValidationError, APIException
from arkindex.documents.models import Corpus, Right
from arkindex.project.elastic import ESQuerySet
from arkindex.project.tools import elasticsearch_escape


class CorpusACLMixin(object):

    def get_corpus(self, corpus_id, right=Right.Read):
        corpus = get_object_or_404(Corpus, id=corpus_id)
        if right not in corpus.get_acl_rights(self.request.user):
            raise PermissionDenied()
        return corpus

    def has_read_access(self, corpus):
        assert isinstance(corpus, Corpus)
        return corpus.public or self.request.user.is_admin or Right.Read in corpus.get_acl_rights(self.request.user)

    def has_write_access(self, corpus):
        assert isinstance(corpus, Corpus)
        return self.request.user.is_admin or not self.request.user.is_anonymous and \
            Right.Write in corpus.get_acl_rights(self.request.user)

    def has_admin_access(self, corpus):
        assert isinstance(corpus, Corpus)
        return self.request.user.is_admin or not self.request.user.is_anonymous and \
            Right.Admin in corpus.get_acl_rights(self.request.user)


class SearchAPIMixin(CorpusACLMixin):

    def get(self, request, *args, **kwargs):
        q = request.query_params.get('q')
        if not q or q.isspace():
            raise ValidationError('A search query is required')
        return super().get(request, *args, **kwargs)

    def get_context(self):
        context = {
            'query': elasticsearch_escape(self.request.query_params['q']),
            'type': self.request.query_params.get('type'),
            'inner_hits_size': settings.ES_INNER_RESULTS_LIMIT,
        }
        if 'corpus' in self.request.query_params:
            try:
                context['corpus_id'] = str(self.get_corpus(self.request.query_params['corpus']).id)
            except Corpus.DoesNotExist:
                raise PermissionDenied
        else:
            context['corpora_ids'] = Corpus.objects.readable(self.request.user).values_list('id', flat=True)
        return context

    def get_template_path(self):
        if self.template_path:
            return self.template_path
        raise APIException('A JSON template path is required to build an ElasticSearch query.')

    def get_query(self):
        if self.es_query:
            return self.es_query

        return ESQuerySet.make_query(
            self.get_template_path(),
            ctx=self.get_context(),
        )

    def get_sort(self):
        return self.es_sort or {}

    def get_index(self):
        if not self.es_index:
            raise APIException('An ElasticSearch index name is required.')
        return self.es_index

    def get_type(self):
        return self.es_type or '_doc'

    def get_queryset(self):
        return ESQuerySet(
            _source=self.es_source,
            query=self.get_query(),
            sort=self.get_sort(),
            es_index=self.get_index(),
            es_type=self.get_type(),
            post_process=self.post_process,
        )

    def post_process(self, *args):
        return args
